# Part of the ROBOID project - http://hamster.school
# Copyright (C) 2016 Kwang-Hyun Park (akaii@kw.ac.kr)
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General
# Public License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330,
# Boston, MA  02111-1307  USA

import time

from roboid.runner import Runner
from roboid.model import DeviceType
from roboid.model import DataType
from roboid.model import Roboid
from roboid.connector import Result
from roboid.hamster import Hamster
from roboid.serial_connector import SerialConnector


class HamsterRoboid(Roboid):
    def __init__(self, index):
        Roboid.__init__(self, 'Hamster', 0x00400000)
        self.__index = index
        self.__connector = None
        self.__ready = False

        self.__left_wheel = 0
        self.__right_wheel = 0
        self.__buzzer = 0
        self.__output_a = 0
        self.__output_b = 0
        self.__topology = 0
        self.__left_led = 0
        self.__right_led = 0
        self.__note = 0
        self.__line_tracer_mode = 0
        self.__line_tracer_speed = 5
        self.__io_mode_a = 0
        self.__io_mode_b = 0
        self.__config_proximity = 2
        self.__config_gravity = 0
        self.__config_band_width = 3

        self.__light = 0
        self.__temperature = 0
        self.__line_tracer_state = 0
        self.__line_tracer_flag = 0
        self.__line_tracer_event = 0
        self.__line_tracer_mode_written = False

        self.__create_model()

    def get_id(self):
        return Hamster.ID

    def __create_model(self):
        from roboid.hamster import Hamster
        dict = self.__device_dict = {}
        dict[Hamster.LEFT_WHEEL] = self.__left_wheel_device = self._Roboid__add_device(Hamster.LEFT_WHEEL, 'LeftWheel', DeviceType.EFFECTOR, DataType.INTEGER, 1, -100, 100, 0)
        dict[Hamster.RIGHT_WHEEL] = self.__right_wheel_device = self._Roboid__add_device(Hamster.RIGHT_WHEEL, 'RightWheel', DeviceType.EFFECTOR, DataType.INTEGER, 1, -100, 100, 0)
        dict[Hamster.BUZZER] = self.__buzzer_device = self._Roboid__add_device(Hamster.BUZZER, 'Buzzer', DeviceType.EFFECTOR, DataType.FLOAT, 1, 0, 167772.15, 0.0)
        dict[Hamster.OUTPUT_A] = self.__output_a_device = self._Roboid__add_device(Hamster.OUTPUT_A, 'OutputA', DeviceType.EFFECTOR, DataType.INTEGER, 1, 0, 255, 0)
        dict[Hamster.OUTPUT_B] = self.__output_b_device = self._Roboid__add_device(Hamster.OUTPUT_B, 'OutputB', DeviceType.EFFECTOR, DataType.INTEGER, 1, 0, 255, 0)
        dict[Hamster.TOPOLOGY] = self.__topology_device = self._Roboid__add_device(Hamster.TOPOLOGY, 'Topology', DeviceType.COMMAND, DataType.INTEGER, 1, 0, 15, 0)
        dict[Hamster.LEFT_LED] = self.__left_led_device = self._Roboid__add_device(Hamster.LEFT_LED, 'LeftLed', DeviceType.COMMAND, DataType.INTEGER, 1, 0, 7, 0)
        dict[Hamster.RIGHT_LED] = self.__right_led_device = self._Roboid__add_device(Hamster.RIGHT_LED, 'RightLed', DeviceType.COMMAND, DataType.INTEGER, 1, 0, 7, 0)
        dict[Hamster.NOTE] = self.__note_device = self._Roboid__add_device(Hamster.NOTE, 'Note', DeviceType.COMMAND, DataType.INTEGER, 1, 0, 88, 0)
        dict[Hamster.LINE_TRACER_MODE] = self.__line_tracer_mode_device = self._Roboid__add_device(Hamster.LINE_TRACER_MODE, 'LineTracerMode', DeviceType.COMMAND, DataType.INTEGER, 1, 0, 15, 0)
        dict[Hamster.LINE_TRACER_SPEED] = self.__line_tracer_speed_device = self._Roboid__add_device(Hamster.LINE_TRACER_SPEED, 'LineTracerSpeed', DeviceType.COMMAND, DataType.INTEGER, 1, 1, 8, 5)
        dict[Hamster.IO_MODE_A] = self.__io_mode_a_device = self._Roboid__add_device(Hamster.IO_MODE_A, 'IoModeA', DeviceType.COMMAND, DataType.INTEGER, 1, 0, 15, 0)
        dict[Hamster.IO_MODE_B] = self.__io_mode_b_device = self._Roboid__add_device(Hamster.IO_MODE_B, 'IoModeB', DeviceType.COMMAND, DataType.INTEGER, 1, 0, 15, 0)
        dict[Hamster.CONFIG_PROXIMITY] = self.__config_proximity_device = self._Roboid__add_device(Hamster.CONFIG_PROXIMITY, 'ConfigProximity', DeviceType.COMMAND, DataType.INTEGER, 1, 1, 7, 2)
        dict[Hamster.CONFIG_GRAVITY] = self.__config_gravity_device = self._Roboid__add_device(Hamster.CONFIG_GRAVITY, 'ConfigGravity', DeviceType.COMMAND, DataType.INTEGER, 1, 0, 3, 0)
        dict[Hamster.CONFIG_BAND_WIDTH] = self.__config_band_width_device = self._Roboid__add_device(Hamster.CONFIG_BAND_WIDTH, 'ConfigBandWidth', DeviceType.COMMAND, DataType.INTEGER, 1, 1, 8, 3)
        dict[Hamster.SIGNAL_STRENGTH] = self.__signal_strength_device = self._Roboid__add_device(Hamster.SIGNAL_STRENGTH, 'SignalStrength', DeviceType.SENSOR, DataType.INTEGER, 1, -128, 0, 0)
        dict[Hamster.LEFT_PROXIMITY] = self.__left_proximity_device = self._Roboid__add_device(Hamster.LEFT_PROXIMITY, 'LeftProximity', DeviceType.SENSOR, DataType.INTEGER, 1, 0, 255, 0)
        dict[Hamster.RIGHT_PROXIMITY] = self.__right_proximity_device = self._Roboid__add_device(Hamster.RIGHT_PROXIMITY, 'RightProximity', DeviceType.SENSOR, DataType.INTEGER, 1, 0, 255, 0)
        dict[Hamster.LEFT_FLOOR] = self.__left_floor_device = self._Roboid__add_device(Hamster.LEFT_FLOOR, 'LeftFloor', DeviceType.SENSOR, DataType.INTEGER, 1, 0, 255, 0)
        dict[Hamster.RIGHT_FLOOR] = self.__right_floor_device = self._Roboid__add_device(Hamster.RIGHT_FLOOR, 'RightFloor', DeviceType.SENSOR, DataType.INTEGER, 1, 0, 255, 0)
        dict[Hamster.ACCELERATION] = self.__acceleration_device = self._Roboid__add_device(Hamster.ACCELERATION, 'Acceleration', DeviceType.SENSOR, DataType.INTEGER, 3, -32768, 32767, 0)
        dict[Hamster.LIGHT] = self.__light_device = self._Roboid__add_device(Hamster.LIGHT, 'Light', DeviceType.SENSOR, DataType.INTEGER, 1, 0, 65535, 0)
        dict[Hamster.TEMPERATURE] = self.__temperature_device = self._Roboid__add_device(Hamster.TEMPERATURE, 'Temperature', DeviceType.SENSOR, DataType.INTEGER, 1, -40, 88, 0)
        dict[Hamster.INPUT_A] = self.__input_a_device = self._Roboid__add_device(Hamster.INPUT_A, 'InputA', DeviceType.SENSOR, DataType.INTEGER, 1, 0, 255, 0)
        dict[Hamster.INPUT_B] = self.__input_b_device = self._Roboid__add_device(Hamster.INPUT_B, 'InputB', DeviceType.SENSOR, DataType.INTEGER, 1, 0, 255, 0)
        dict[Hamster.LINE_TRACER_STATE] = self.__line_tracer_state_device = self._Roboid__add_device(Hamster.LINE_TRACER_STATE, 'LineTracerState', DeviceType.EVENT, DataType.INTEGER, 1, 0, 255, 0)

    def find_device_by_id(self, device_id):
        return self.__device_dict.get(device_id)

    def __run(self):
        try:
            while self.__running:
                if self.__receive(self.__connector):
                    self.__send(self.__connector)
                time.sleep(0.01)
        except:
            pass

    def __init(self, port_name=None):
        import threading
        Runner.register_required()
        self.__running = True
        thread = threading.Thread(target=self.__run)
        self.__thread = thread
        thread.daemon = True
        thread.start()

        tag = 'Hamster[{}]'.format(self.__index)
        self.__connector = SerialConnector(tag, 'Hamster', '04')
        result = self.__connector.open(port_name)
        if result == Result.FOUND:
            while self.__ready == False and self._Roboid__is_disposed() == False:
                time.sleep(0.01)
        elif result == Result.NOT_AVAILABLE:
            Runner.register_checked()

    def __release(self):
        self.__reset()
        self.__running = False
        thread = self.__thread
        self.__thread = None
        if thread:
            thread.join()

        connector = self.__connector
        self.__connector = None
        if connector:
            connector.close()

    def __dispose(self):
        if self._Roboid__is_disposed() == False:
            self._Roboid__dispose()
            self.__release()

    def __reset(self):
        self._Roboid__reset()

        self.__left_wheel = 0
        self.__right_wheel = 0
        self.__buzzer = 0
        self.__output_a = 0
        self.__output_b = 0
        self.__topology = 0
        self.__left_led = 0
        self.__right_led = 0
        self.__note = 0
        self.__line_tracer_mode = 0
        self.__line_tracer_speed = 5
        self.__io_mode_a = 0
        self.__io_mode_b = 0
        self.__config_proximity = 2
        self.__config_gravity = 0
        self.__config_band_width = 3

        self.__light = 0
        self.__temperature = 0
        self.__line_tracer_state = 0
        self.__line_tracer_flag = 0
        self.__line_tracer_event = 0
        self.__line_tracer_mode_written = False

    def __to_hex(self, number):
        value = int(number)
        if value < 0: value += 0x100
        value = hex(value).upper()[2:]
        if len(value) > 1:
            return value
        return '0' + value

    def __to_hex3(self, number):
        value = int(number)
        if value < 0: value += 0x1000000
        value = hex(value).upper()[2:]
        length = len(value)
        if length < 6:
            result = ''
            for i in range(length, 6):
                result += '0'
            return result + value
        return value

    def __request_motoring_data(self):
        self.__left_wheel = self.__left_wheel_device.read()
        self.__right_wheel = self.__right_wheel_device.read()
        self.__buzzer = self.__buzzer_device.read()
        self.__output_a = self.__output_a_device.read()
        self.__output_b = self.__output_b_device.read()
        if self.__topology_device._Device__is_written():
            self.__topology = self.__topology_device.read()
        if self.__left_led_device._Device__is_written():
            self.__left_led = self.__left_led_device.read()
        if self.__right_led_device._Device__is_written():
            self.__right_led = self.__right_led_device.read()
        if self.__note_device._Device__is_written():
            self.__note = self.__note_device.read()
        if self.__line_tracer_mode_device._Device__is_written():
            self.__line_tracer_mode = self.__line_tracer_mode_device.read()
            self.__line_tracer_mode_written = True
        if self.__line_tracer_speed_device._Device__is_written():
            self.__line_tracer_speed = self.__line_tracer_speed_device.read()
        if self.__io_mode_a_device._Device__is_written():
            self.__io_mode_a = self.__io_mode_a_device.read()
        if self.__io_mode_b_device._Device__is_written():
            self.__io_mode_b = self.__io_mode_b_device.read()
        if self.__config_proximity_device._Device__is_written():
            self.__config_proximity = self.__config_proximity_device.read()
        if self.__config_gravity_device._Device__is_written():
            self.__config_gravity = self.__config_gravity_device.read()
        if self.__config_band_width_device._Device__is_written():
            self.__config_band_width = self.__config_band_width_device.read()
        self._Roboid__clear_written()

    def __encode_motoring_packet(self, address):
        result = ''
        result += self.__to_hex(self.__topology & 0x0f)
        result += '0010'
        result += self.__to_hex(self.__left_wheel)
        result += self.__to_hex(self.__right_wheel)
        result += self.__to_hex(self.__left_led)
        result += self.__to_hex(self.__right_led)
        result += self.__to_hex3(self.__buzzer * 100)
        result += self.__to_hex(self.__note)
        if self.__line_tracer_mode_written:
            if self.__line_tracer_mode > 0:
                self.__line_tracer_flag ^= 0x80
                self.__line_tracer_event = 1
            self.__line_tracer_mode_written = False
        temp = (self.__line_tracer_mode & 0x0f) << 3
        temp |= (self.__line_tracer_speed - 1) & 0x07
        temp |= self.__line_tracer_flag
        result += self.__to_hex(temp)
        result += self.__to_hex(self.__config_proximity)
        temp = (self.__config_gravity & 0x0f) << 4
        temp |= self.__config_band_width & 0x0f
        result += self.__to_hex(temp)
        temp = (self.__io_mode_a & 0x0f) << 4
        temp |= self.__io_mode_b & 0x0f
        result += self.__to_hex(temp)
        result += self.__to_hex(self.__output_a)
        result += self.__to_hex(self.__output_b)
        result += '000000-'
        result += address
        result += '\r'
        return result

    def __decode_sensory_packet(self, packet):
        packet = str(packet)
        value = int(packet[6:8], 16)
        value -= 0x100
        self.__signal_strength_device._Device__put(value)
        value = int(packet[8:10], 16)
        self.__left_proximity_device._Device__put(value)
        value = int(packet[10:12], 16)
        self.__right_proximity_device._Device__put(value)
        value = int(packet[12:14], 16)
        self.__left_floor_device._Device__put(value)
        value = int(packet[14:16], 16)
        self.__right_floor_device._Device__put(value)
        value = int(packet[16:20], 16)
        if value > 0x7fff: value -= 0x10000
        self.__acceleration_device._Device__put_at(0, value)
        value = int(packet[20:24], 16)
        if value > 0x7fff: value -= 0x10000
        self.__acceleration_device._Device__put_at(1, value)
        value = int(packet[24:28], 16)
        if value > 0x7fff: value -= 0x10000
        self.__acceleration_device._Device__put_at(2, value)
        value = int(packet[28:30], 16)
        if value == 0:
            self.__light = int(packet[30:34], 16)
        else:
            value = int(packet[30:32], 16)
            if value > 0x7f: value -= 0x100
            self.__temperature = int(value / 2.0 + 24)
        self.__light_device._Device__put(self.__light)
        self.__temperature_device._Device__put(self.__temperature)
        value = int(packet[34:36], 16)
        self.__input_a_device._Device__put(value)
        value = int(packet[36:38], 16)
        self.__input_b_device._Device__put(value)
        value = int(packet[38:40], 16)
        if (value & 0x40) != 0:
            if self.__line_tracer_event == 1:
                if value != 0x40:
                    self.__line_tracer_event = 2
            if self.__line_tracer_event == 2:
                if value != self.__line_tracer_state:
                    self.__line_tracer_state = value
                    self.__line_tracer_state_device._Device__put(value)
                    if value == 0x40:
                        self.__line_tracer_event = 0
        return True

    def __receive(self, connector):
        if connector:
            packet = connector.read()
            if packet:
                if self.__decode_sensory_packet(packet):
                    if self.__ready == False:
                        self.__ready = True
                        Runner.register_checked()
                    self._Roboid__notify_sensory_device_data_changed()
                return True
        return False

    def __send(self, connector):
        if connector:
            packet = self.__encode_motoring_packet(connector.get_address())
            connector.write(packet)