# Part of the ROBOID project - http://hamster.school
# Copyright (C) 2016 Kwang-Hyun Park (akaii@kw.ac.kr)
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# 
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General
# Public License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330,
# Boston, MA  02111-1307  USA

import sys
from timeit import default_timer as timer

import serial.tools.list_ports

from roboid.connector import State
from roboid.connector import Result


BAUD_RATE = 115200
VALID_PACKET_LENGTH = 54
RETRY = 10
TIMEOUT = 0.1


class SerialConnector:
    def __init__(self, tag, connector_name, connector_code):
        self.__tag = tag
        self.__connector_name = connector_name
        self.__connector_code = connector_code
        self.__serial = None
        self.__address = '000000000000'
        self.__port_name = ''
        self.__found = False
        self.__timestamp = 0
        self.__connected = False

    def open(self, port_name=None):
        if port_name:
            result = self.__open_port(port_name)
            if result != Result.NOT_AVAILABLE:
                return result
        else:
            ports = serial.tools.list_ports.comports()
            for port in ports:
                result = self.__open_port(port[0])
                if result != Result.NOT_AVAILABLE:
                    return result
        self.__print_error('No available USB to BLE bridge')
        return Result.NOT_AVAILABLE

    def __open_port(self, port_name):
        if port_name:
            try:
                s = serial.Serial(port_name, BAUD_RATE, rtscts=True, timeout=0.1)
                s.reset_input_buffer()
                s.reset_output_buffer()
                self.__port_name = port_name
                result = self.__check_port(s)
                if result != Result.NOT_AVAILABLE:
                    self.__serial = s
                    return result
                s.close()
            except:
                pass
        return Result.NOT_AVAILABLE

    def close(self):
        self.__connected = False
        if self.__serial:
            self.__serial.close()
            self.__serial = None
        self.__print_message('Disposed')

    def is_connected(self):
        return self.__connected

    def get_address(self):
        return self.__address

    def __set_address(self, address):
        self.__address = address

    def __set_connection_state(self, state):
        self.__connected = (state == State.CONNECTED)
        if self.__found == False and self.__connected:
            self.__found = True
        if self.__found:
            if state == State.CONNECTED:
                address = self.__address
                if len(address) >= 12:
                    self.__print_message('Connected: {} {}:{}:{}:{}:{}:{}'.format(self.__port_name, address[10:12], address[8:10], address[6:8], address[4:6], address[2:4], address[0:2]))
                else:
                    self.__print_message('Connected: {}'.format(self.__port_name))
            elif state == State.CONNECTION_LOST:
                self.__print_error('Connection lost')

    def __read_packet(self, serial, start_byte=None):
        try:
            packet = serial.read_until('\r')
            if start_byte == None:
                return packet
            if packet[:2] == start_byte:
                return packet
            return None
        except:
            return None

    def write(self, packet):
        if self.__serial:
            try:
                self.__serial.write(packet)
            except:
                pass

    def read(self):
        if self.__serial:
            try:
                packet = self.__serial.read_until('\r')
                if len(packet) == VALID_PACKET_LENGTH:
                    if self.__found == False:
                        self.__check_connection(self.__serial)
                    elif self.__connected == False:
                        self.__set_connection_state(State.CONNECTED)
                    self.__timestamp = 0
                    return packet
                elif self.__connected:
                    t = timer()
                    if self.__timestamp == 0:
                        self.__timestamp = t
                    elif t - self.__timestamp > TIMEOUT:
                        self.__set_connection_state(State.CONNECTION_LOST)
            except:
                if self.__connected:
                    self.__set_connection_state(State.CONNECTION_LOST)
        return None

    def __check_port(self, serial):
        self.__read_packet(serial)
        packet1 = self.__read_packet(serial)
        packet2 = self.__read_packet(serial)
        if packet2:
            if len(packet2) == VALID_PACKET_LENGTH:
                return self.__check_connection(serial)
            elif packet1 and len(packet2) == 2:
                self.__print_error('Not connected')
                return Result.NOT_CONNECTED
        return Result.NOT_AVAILABLE

    def __check_connection(self, serial):
        for i in range(RETRY):
            serial.write('FF\r')
            packet = self.__read_packet(serial, 'FF')
            if packet:
                packet = packet.strip()
                info = packet.split(',')
                if info and len(info) >= 5:
                    if info[1] == self.__connector_name and info[2] == self.__connector_code:
                        self.__set_address(info[4])
                        self.__set_connection_state(State.CONNECTED)
                        return Result.FOUND
                return Result.NOT_AVAILABLE
        return Result.NOT_AVAILABLE

    def __print_message(self, message):
        sys.stdout.write('{} {}\n'.format(self.__tag, message))

    def __print_error(self, message):
        sys.stderr.write('{} {}\n'.format(self.__tag, message))